# Databricks notebook source exported at Wed, 23 Nov 2016 15:15:25 UTC
# MAGIC  %md 
# MAGIC Introduction
# MAGIC ============
# MAGIC  
# MAGIC This PySpark notebook introduces Spark GraphFrames.  
# MAGIC  
# MAGIC The dataset
# MAGIC -----------
# MAGIC  
# MAGIC SNAP: https://snap.stanford.edu/data/egonets-Facebook.html
# MAGIC 
# MAGIC Dataset description:
# MAGIC * Nodes	4039
# MAGIC * Edges	88234
# MAGIC 
# MAGIC Abstract from Stanford's website: 
# MAGIC This dataset consists of 'circles' (or 'friends lists') from Facebook. Facebook data was collected from survey participants using a Facebook app. The dataset includes node features (profiles), circles, and ego networks.
# MAGIC 
# MAGIC Facebook data has been anonymized by replacing the Facebook-internal ids for each user with a new value. Also, while feature vectors from this dataset have been provided, the interpretation of those features has been obscured. For instance, where the original dataset may have contained a feature "political=Democratic Party", the new data would simply contain "political=anonymized feature 1". Thus, using the anonymized data it is possible to determine whether two users have the same political affiliations, but not what their individual political affiliations represent.
# MAGIC 
# MAGIC For this notebook, the data files has been downloaded, cleaned from duplicate data and properly formatted in csv format for better handling. You can download the vertices file from [here](http://www.clounce.com/wp-content/uploads/2016/11/stanford_fb_vertices.csv) and the edges from [here](http://www.clounce.com/wp-content/uploads/2016/11/stanford_fb_edges.csv).  
# MAGIC 
# MAGIC All files, including this notebook and draft D3js html code can be downloaded from [here](http://www.clounce.com/wp-content/uploads/2016/11/snap.zip).  Thanks!

# COMMAND ----------

# files containing vertices and edges, these are donwloaded from SNAP website, cleaned, and formatted for our use
# files can be accessed from https://community.cloud.databricks.com/files/tables/fovepx7h1479410900486/vertices.csv
# create new RDDs from each file
e = sc.textFile("/FileStore/tables/fovepx7h1479410900486/edges.csv")
v = sc.textFile("/FileStore/tables/fovepx7h1479410900486/vertices.csv")

print("Number of partitions for e: %d" % e.getNumPartitions())
print("Number of partitions for v: %d" % v.getNumPartitions())

# COMMAND ----------

# MAGIC %md
# MAGIC ### RDD (Resilient Distributed Datasets)
# MAGIC 
# MAGIC * It is the __building block__ of spark. All data abstractions, such as DataFrames and GraphFrames, are interprested (transformed) in RDDs.
# MAGIC * RDD is __lazily evaluated immutable parallel collection of objects__ normally exposed with lambda functions.
# MAGIC * RDDs are __simple to use__ and expose an Object Oriented like API. See [Spark Programming Guide](http://spark.apache.org/docs/latest/programming-guide.html#resilient-distributed-datasets-rdds).
# MAGIC * Its main __disadvantage is performance limitations__. Being in-memory JVM objects, RDDs involve overhead of Garbage Collection and Java Serialisation which become expensive when data grows.

# COMMAND ----------

# let's view what's in the vertices RDD
v.collect()

# COMMAND ----------

# helper function to convert from string value to integer
def convertToInt(s):
  if s == None or s == 'None':
    return None
  else:
    return int(s)

# COMMAND ----------

# the uploaded files have header and data is stored as string
# we need to remove the header and convert the strings to integers; None fields will be translated to null
eheader = e.first() #extract edge header
edges = e.filter(lambda row:row != eheader and row != "").map(lambda line:line.split(",")).map(lambda line:(convertToInt(line[0]), convertToInt(line[1])))

vheader = v.first() #extract vertex header
vertices = v.filter(lambda row:row != vheader).map(lambda line:line.split(",")).map(lambda line:(convertToInt(line[0]), convertToInt(line[1]), convertToInt(line[2]), convertToInt(line[3]), convertToInt(line[4]), convertToInt(line[5])))

print("Vertex data: %s" % vheader)

# COMMAND ----------

# vertices RDD after map function
vertices.collect()

# COMMAND ----------

# MAGIC %md
# MAGIC ### DataFrames
# MAGIC 
# MAGIC A DataFrame is a Dataset organized into named columns. It is conceptually equivalent to a table in a relational database or a data frame in R/Python.  See (Spark SQL, DataFrames and Datasets Guide)[http://spark.apache.org/docs/latest/sql-programming-guide.html].

# COMMAND ----------

# Next we will convert our RDDs to DataFrames
# We will first create the schema and then create the dataframe
from pyspark.sql.types import *

# GraphFrames will expect to have our key named as 'id'
vertexSchema = StructType([StructField("id", IntegerType(), False),
                           StructField("birthday", IntegerType(), True),
                           StructField("hometown_id", IntegerType(), True),
                           StructField("work_employer_id", IntegerType(), True),
                           StructField("education_school_id", IntegerType(), True),
                           StructField("education_year_id", IntegerType(), True)])

# Create a data frame
vdf = sqlContext.createDataFrame(vertices, vertexSchema)

# COMMAND ----------

# Let's look at the contents of the dataframe
display(vdf)

# COMMAND ----------

# We will do the same for edges.
# Note that later on GraphFrames require 'src' and 'dst' headers
edgeSchema = StructType([StructField("src", IntegerType(), False),
                           StructField("dst", IntegerType(), False)])

# Create a data frame
edf = sqlContext.createDataFrame(edges, edgeSchema)

# COMMAND ----------

# MAGIC %md
# MAGIC 
# MAGIC ### GraphFrames
# MAGIC 
# MAGIC GraphFrames is a package for Apache Spark which provides DataFrame-based Graphs. It provides high-level APIs in Scala, Java, and Python. It aims to provide both the functionality of GraphX and extended functionality taking advantage of Spark DataFrames. This extended functionality includes motif finding, DataFrame-based serialization, and highly expressive graph queries.
# MAGIC 
# MAGIC See [GraphFrames documentation](http://graphframes.github.io/index.html).

# COMMAND ----------

# Once we have our two DataFrames, we will now create the GraphFrame
from graphframes import *

g = GraphFrame(vdf, edf)

# COMMAND ----------

display(g.edges)

# COMMAND ----------

display(g.outDegrees)

# COMMAND ----------

# find all connected vertices with the same birthday identifier
print "same birthdays"
res = g.find("(a)-[]->(b)") \
         .filter("a.birthday = b.birthday")
print "count: %d" % res.count()
display(res.select("a.id", "b.id", "b.birthday"))

# COMMAND ----------

from pyspark.sql.functions import *

# COMMAND ----------

# We will now export the birthdays result to csv so that we display it using D3.js

# clean filesystem from previous runs
dbutils.fs.rm("/FileStore/tables/fovepx7h1479410900486/birthdays.csv", True)

res.select(col("a.id").alias("source"), col("b.id").alias("target")).coalesce(1).write.format("com.databricks.spark.csv").save("/FileStore/tables/fovepx7h1479410900486/birthdays.csv")

# COMMAND ----------

# check the file path
# got to https://community.cloud.databricks.com/files/tables/fovepx7h1479410900486/birthdays.csv/....csv
# from bowser tools, copy/paste list of edges

# uncomment the line below to view the result file
# dbutils.fs.ls("/FileStore/tables/fovepx7h1479410900486/birthdays.csv")

# COMMAND ----------

# find "friends of friends" who are not connected to us, but graduated the same
# year from the same school
print "same class"
res = g.find("(a)-[]->(b); (b)-[]->(c); !(a)-[]->(c)") \
         .filter("a.education_school_id = c.education_school_id and " \
                 "a.education_year_id = c.education_year_id")
res = res.filter("a.id != c.id").select(col("a.id").alias("source"), "a.education_school_id", "a.education_year_id",
        col("c.id").alias("target"), "c.education_school_id", "c.education_year_id") 
print "count: %d" % res.count()

display(res)

# COMMAND ----------

# clean filesystem from previous runs
dbutils.fs.rm("/FileStore/tables/fovepx7h1479410900486/fofg.csv", True)

res.select("source", "target").coalesce(1).write.format("com.databricks.spark.csv").save("/FileStore/tables/fovepx7h1479410900486/fofg.csv")

# COMMAND ----------

# uncomment the line below to view the result file
# dbutils.fs.ls("/FileStore/tables/fovepx7h1479410900486/fofg.csv")

# COMMAND ----------

# do a page rank on the graph
print "page rank"
pgr = g.pageRank(resetProbability=0.15, tol=0.01).vertices.sort(
    	    'pagerank', ascending=False)
display(pgr.select("id", "pagerank"))

# COMMAND ----------

print "links from 1911 - highest rank"
res = g.find("(a)-[]->(b); (b)-[]->(c)").filter("c.id = 1911")
display(res)

# COMMAND ----------

# we will now create two files with (a)-[]->(b) and (b)-[]->(c) and then concatenate them for display in D3js
highrankL1 = res.select(col("a.id").alias("source"), col("b.id").alias("target"))
highrankL2 = res.select(col("b.id").alias("source"), col("c.id").alias("target"))

# COMMAND ----------

# clean filesystem from previous runs
dbutils.fs.rm("/FileStore/tables/fovepx7h1479410900486/highrankL1.csv", True)
dbutils.fs.rm("/FileStore/tables/fovepx7h1479410900486/highrankL2.csv", True)

highrankL1.select("source", "target").coalesce(1).write.csv("/FileStore/tables/fovepx7h1479410900486/highrankL1.csv")
highrankL2.select("source", "target").coalesce(1).write.csv("/FileStore/tables/fovepx7h1479410900486/highrankL2.csv")

# COMMAND ----------

# uncomment the line below to view the files in HDFS
# dbutils.fs.ls("/FileStore/tables/fovepx7h1479410900486/highrankL1.csv")

# COMMAND ----------

# uncomment the line below to view the files in HDFS
# dbutils.fs.ls("/FileStore/tables/fovepx7h1479410900486/highrankL2.csv")

# COMMAND ----------


